/*
 * Copyright 2024 Amazon.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  http://aws.amazon.com/apache2.0
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package software.amazon.smithy.go.codegen.trait;

import static software.amazon.smithy.model.node.Node.objectNode;

import software.amazon.smithy.model.shapes.ShapeId;
import software.amazon.smithy.model.traits.AnnotationTrait;

/**
 * Marker trait to indicate that an input or output shape on an operation was backfilled (i.e. the operation DID NOT
 * have a target for it in the original model).
 */
public class BackfilledInputOutputTrait extends AnnotationTrait {
    public static final ShapeId ID = ShapeId.from("smithy.go.trait#BackfilledInputOutput");

    public BackfilledInputOutputTrait() {
        super(ID, objectNode());
    }
}
