// package xml is xml testing package that supports xml comparison utility.
// The package consists of XMLToStruct and StructToXML utils that help sort xml elements
// as per their nesting level. XMLToStruct function converts an xml document into a sorted
// tree node structure, while StructToXML converts the sorted xml nodes into a sorted xml document.
// SortXML function should be used to sort an xml document. It can be configured to ignore indentation
package xml
