/*
 * Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  http://aws.amazon.com/apache2.0
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package software.amazon.smithy.go.codegen.integration;

/**
 * Represents the middleware stack step.
 */
public enum MiddlewareStackStep {
    INITIALIZE,
    BUILD,
    SERIALIZE,
    DESERIALIZE,
    FINALIZE;

    @Override
    public String toString() {
        switch (this) {
            case INITIALIZE:
                return "Initialize";
            case BUILD:
                return "Build";
            case SERIALIZE:
                return "Serialize";
            case DESERIALIZE:
                return "Deserialize";
            case FINALIZE:
                return "Finalize";
            default:
                return "Unknown";
        }
    }
}
